potool jest programem wspomagającym edycję plików po
Copyright (C) 1999-2003 Zbigniew Chyla
Copyright (C) 2000-2012 Marcin Owsiany <porridge@debian.org>

Licencja znajduje się w pliku LICENSE

Pakiety potrzebne do kompilacji programu:
	gcc, bison, flex, glib 2.x w wersji -devel

Dokumentacja dostępna jest w postaci stron podręcznika (man):
  poedit.pl.1
  potool.pl.1
  postats.pl.1
  change-po-charset.pl.1
