/*
 * potool is a program aiding editing of po files
 * Copyright (C) 1999-2002 Zbigniew Chyla
 *
 * see LICENSE for licensing info
 */
#ifndef I18N_H
#define I18N_H

#define _(string)  (string)

#endif
